# -*- coding: utf-8 -*-

from base64 import b64encode
import datetime
import logging
import os
import sys

if sys.platform == 'win32':
    import pbs as sh
else:
    import sh

from openerp.osv import fields, osv
from openerp.tools import DEFAULT_SERVER_DATE_FORMAT
from openerp.tools import DEFAULT_SERVER_DATETIME_FORMAT
from openerp.tools import config
from openerp.tools.translate import _

from .util import get_secure_filename

_logger = logging.getLogger(__name__)


class job_log(osv.Model):
    _name = 'external_job.job_log'

    def _get_filename(self, cr, uid, ids, field_name, arg, context=None):
        attachment_osv = self.pool['ir.attachment']
        search_args = [
            '&',
            ('res_model', '=', 'external_job.job_log'),
            ('res_id', 'in', ids),
        ]
        attachement_ids = attachment_osv.search(
            cr, uid, search_args, context=context)
        result = dict.fromkeys(ids)
        read_args = ['datas', 'datas_fname', 'res_id']
        attachments = attachment_osv.read(
            cr, uid, attachement_ids, read_args, context=context)
        for attachment in attachments:
            result[attachment.get('res_id')] = attachment.get('datas_fname')

        return result

    def _get_attachment(self, cr, uid, ids, field_name, arg, context=None):
        attachment_osv = self.pool['ir.attachment']
        search_args = [
            '&',
            ('res_model', '=', 'external_job.job_log'),
            ('res_id', 'in', ids),
        ]
        attachement_ids = attachment_osv.search(
            cr, uid, search_args, context=context)
        result = dict.fromkeys(ids)
        read_args = ['datas', 'datas_fname', 'res_id']
        attachments = attachment_osv.read(
            cr, uid, attachement_ids, read_args, context=context)
        for attachment in attachments:
            result[attachment.get('res_id')] = attachment.get('datas')

        return result

    _columns = {
        'job_definition_id': fields.many2one(
            'external_job.job_definition',
            u"Job Definition",
            ondelete='cascade',
            required=True,
        ),

        'start_date': fields.datetime(
            u"Start Date",
            readonly=True,
        ),
        'end_date': fields.datetime(
            u"End Date",
            readonly=True,
        ),

        'in_file_name': fields.char(
            u"In File Name",
            size=512,
            readonly=True,
        ),
        'out_file_name': fields.char(
            u"Out File Name",
            size=512,
            readonly=True,
        ),

        'out_file': fields.function(
            _get_attachment,
            type='binary',
            string=u"Output File",
        ),
        'filename': fields.function(
            _get_filename,
            type='char',
            string=u"Filename",
        ),

        'stdout': fields.text(
            u"StdOut",
            readonly=True,
        ),
        'stderr': fields.text(
            u"StdErr",
            readonly=True,
        ),
    }

    _defaults = {
        'start_date': lambda *a: fields.datetime.now(),
    }

    def name_get(self, cr, uid, ids, context=None):
        # TODO There might be a method that grab the lang and timezone from
        # the context and format a datetime correctly.
        res = []
        if context and 'lang' in context:
            language_osv = self.pool['res.lang']
            lang_id = language_osv.search(
                cr, uid, [('code', '=', context['lang'])], context=context)
            lang_r = language_osv.read(
                cr, uid, lang_id, ['date_format', 'time_format'])[0]
            lang_format = lang_r['date_format'] + ' ' + lang_r['time_format']
        else:
            lang_format = '%Y-%m-%d %H:%M:%S'
        for r in self.read(cr, uid, ids, ['start_date']):
            job_datetime = datetime.datetime.strptime(
                r['start_date'], DEFAULT_SERVER_DATETIME_FORMAT)
            job_datetime = fields.datetime.context_timestamp(
                cr, uid, job_datetime, context)
            job_datetime = datetime.datetime.strftime(
                job_datetime, lang_format)
            res.append((r['id'], _(u"%s’s Job") % job_datetime))
        return res

    def run_jobs(self, cr, uid, context=None):
        """Run remaining jobs."""
        # TODO Throttle to allow polling (multiple runners at the same time)?
        job_log_ids = self.search(
            cr, uid,
            [('end_date', '=', False)],
            context=context
        )

        job_logs = self.browse(cr, uid, job_log_ids, context=context)
        for job_log in job_logs:
            self._run_job(cr, uid, job_log, context)

        return True

    def _run_job(self, cr, uid, job_log, context):
        if context is None:
            context = {}

        job_log_obj = self.pool['external_job.job_log']

        job_def = job_log.job_definition_id

        if not job_def.command_line:
            _logger.warning(_(u"No command line for given job"))
            return False
        command = job_def.command_line

        in_file_name = get_secure_filename()
        out_file_name = get_secure_filename()

        if context.get('resource_ids', None):
            # if the caller has injected specific resources ids in the context
            # we will dump them one by one in the input file and give this
            # to the joblog definition so that the job runner (ie: the external
            # script) can get to them and do whatever it needs with it.
            # This mechanism is usefull for specialized commands when we want
            # to create context menus in some list views and extract only the
            # selected ids by passing them in the context
            resource_ids = context.get('resource_ids')
            with open(in_file_name, 'w+') as f:
                for resource_id in resource_ids:
                    f.write("%s\n" % resource_id)

        write_args = {
            'in_file_name': in_file_name,
            'out_file_name': out_file_name,
            'start_date': fields.datetime.now(),
        }
        # Before running the job, safe file names.
        job_log_obj.write(cr, uid, [job_log.id], write_args, context=context)

        # Make sure we commit our transaction before giving the newly created
        # id to our external process or it won't be able to find it :)
        cr.commit()

        args = [cr.dbname, str(job_log.id)]
        if job_def.arguments:
            for argument in job_def.arguments.split():
                args.append(argument)

        if job_def.config_arguments:
            try:
                args.extend(
                    config[key] for key in job_def.config_arguments.split()
                )
            except KeyError as e:
                raise osv.except_osv(
                    _(u"Error"),
                    _(u"Key not found in configuration: {}".format(str(e)))
                )

        try:
            run = sh.Command(command)
        except sh.CommandNotFound:
            raise osv.except_osv(
                _("Error"),
                _("Command %s not found") % command)

        try:
            out = run(*args)
        except sh.ErrorReturnCode as e:
            errcode = _(u"External job {} ({}) failed with error code {}.\n\n")
            raise osv.except_osv(
                _(u"Error"),
                (errcode + u"STDOUT:\n{}\n\n STDERR:\n{}\n\n").format(
                    job_log.id, job_def.name, e.exit_code, e.stdout, e.stderr
                )
            )

        with open(out_file_name, 'rb') as outfile:
            outfile_data = outfile.read()

            if not outfile_data:
                outfile_data = ''

        outfile_data = b64encode(outfile_data)

        write_args = {
            'end_date': fields.datetime.now(),
            'stdout': out,
        }
        job_log_obj.write(cr, uid, job_log.id, write_args, context=context)

        attachment_osv = self.pool['ir.attachment']

        fname = "%s-%s.%s" % (
            job_def.name,
            datetime.datetime.now().date().strftime(
                DEFAULT_SERVER_DATE_FORMAT),
            job_def.file_extension,
        )

        # cleanup temp
        os.remove(in_file_name)
        os.remove(out_file_name)

        create_args = {
            'name': fname,
            'datas': outfile_data,
            'datas_fname': fname,
            'res_id': job_log.id,
            'res_model': 'external_job.job_log',
        }
        attachment_osv.create(cr, uid, create_args, context=context)
